unit OpenFolderDialog;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ShlObj;

type
  TOpenFolderFlags = (
    flBrowseForComputer, // Only returns computers. If the user selects anything other than a computer, the OK button is grayed.
    flBrowseForPrinter,  // Only returns printers. If the user selects anything other than a printer, the OK button is grayed.
    flBrowseForFiles,    //
    flDontGoBelowDomain, // Does not include network folders below the domain level in the tree view control.
    flReturnFSAncestors, // Only returns file system ancestors. If the user selects anything other than a file system ancestor, the OK button is grayed.
    flReturnOnlyFSDirs,  // Only returns file system directories. If the user selects folders that are not part of the file system, the OK button is grayed.
    flShowStatusText     // Includes a status area in the dialog box. The callback function can set the status text by sending messages to the dialog box.
  );
  TOpenFolderOptions = set of TOpenFolderFlags;

  TOpenFolderDialog = class(TCommonDialog)
  private
    { Private declarations }
    BrowseInfo : TBrowseInfo;
    Active : Boolean;
    Handle : HWND;
//    FRootPIDL :
    FDisplayName : String;

    FFolder : String;
    FStatusText : String;

    FTitle : String;
    FOptions : TOpenFolderOptions;
    FOnSelectionChange : TNotifyEvent;
  protected
    { Protected declarations }
    procedure CallBack(Wnd: HWND; uMsg: UINT; lParam : LPARAM);
    procedure SetFolder(Value : String); virtual;
    procedure SetStatusText(Value : String); virtual;
    procedure SetOKButtonEnabled(Value : Boolean); virtual;
  public
    { Public declarations }
    constructor Create(AOwner : TComponent); override;
    function Execute : Boolean; override;
    property OKButtonEnabled : Boolean write SetOKButtonEnabled;
  published
    { Published declarations }
//    property RootPIDL
    property DisplayName : String read FDisplayName;
    property Folder : String read FFolder write SetFolder;
    property StatusText : String read FStatusText write SetStatusText;
    property Title : String read FTitle write FTitle;
    property Options : TOpenFolderOptions read FOptions write FOptions;
    property OnSelectionChange : TNotifyEvent read FOnSelectionChange write FOnSelectionChange;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Dialogs', [TOpenFolderDialog]);
end;

var
  EnstantiationList : TList = Nil;

function SysCallBack(Wnd: HWND; uMsg: UINT; lParam, lpData: LPARAM): Integer; stdcall;
begin
  TOpenFolderDialog(EnstantiationList[lpData]).CallBack(Wnd,uMsg,lParam);
  Result := 0;
end;

constructor TOpenFolderDialog.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  Active := False;
  FOptions := [flReturnOnlyFSDirs];
end;

function GetBIF(Options : TOpenFolderOptions) : Integer;
begin
  Result := 0;
  if flBrowseForComputer in Options then Result := Result or BIF_BROWSEFORCOMPUTER;
  if flBrowseForPrinter in Options then Result := Result or BIF_BROWSEFORPRINTER;
  if flBrowseForFiles in Options then Result := Result or BIF_BROWSEINCLUDEFILES;
  if flDontGoBelowDomain in Options then Result := Result or BIF_DONTGOBELOWDOMAIN;
  if flReturnFSAncestors in Options then Result := Result or BIF_RETURNFSANCESTORS;
  if flReturnOnlyFSDirs in Options then Result := Result or BIF_RETURNONLYFSDIRS;
  if flShowStatusText in Options then Result := Result or BIF_STATUSTEXT;
end;

function TOpenFolderDialog.Execute : Boolean;
var
  Res : PItemIDList;
  ReturnFolder : Array[0..MAX_PATH-1] of Char;
  EnstatiationIndex : Integer;
begin
  EnstatiationIndex := EnstantiationList.Add(Self);

  with BrowseInfo do
    begin
      hwndOwner := 0; //Owner.Handle;
      pidlRoot := Nil;
      pszDisplayName := ReturnFolder;
      lpszTitle := PChar(FTitle);
      ulFlags := GetBIF(FOptions);
      lpfn := SysCallBack;
      lParam := EnstatiationIndex;
    end;

  Active := True;
  Res := SHBrowseForFolder(BrowseInfo);
  Active := False;
  If Assigned(OnClose) then OnClose(Self);
  Result := Res <> Nil;
end;

procedure TOpenFolderDialog.CallBack(Wnd: HWND; uMsg: UINT; lParam : LPARAM);
var
  ReturnFolder : Array[0..MAX_PATH-1] of Char;
begin
  Handle := Wnd;
  case uMsg of
    BFFM_INITIALIZED:
      begin
        if FFolder <> '' then SetFolder(FFolder);
        if FStatusText <> '' then SetStatusText(FStatusText);

        if Assigned(OnShow) then OnShow(Self);
      end;
    BFFM_SELCHANGED:
      begin
        if SHGetPathFromIDList(PItemIDList(lParam),ReturnFolder) then
          FFolder := ReturnFolder;

        if Assigned(FOnSelectionChange) then FOnSelectionChange(Self);
      end;
  end;
end;

procedure TOpenFolderDialog.SetFolder(Value : String);
begin
  if Active then
    SendMessage(Handle, BFFM_SETSELECTION, 1, Integer(PChar(Value)));
  FFolder := Value;
end;

procedure TOpenFolderDialog.SetStatusText(Value : String);
begin
  if Active then
    SendMessage(Handle, BFFM_SETSTATUSTEXT, 1, Integer(PChar(Value)));
  FStatusText := Value;
end;

procedure TOpenFolderDialog.SetOKButtonEnabled(Value : Boolean);
begin
  if not Active then
    raise Exception.Create('OK button status cannot change until the dialog is executing');
  SendMessage(Handle, BFFM_ENABLEOK, Ord(Value), 0);
end;

//var
//  I : Integer;

initialization
  EnstantiationList := TList.Create;
finalization
  //for I := 0 to EnstantiationList.Count-1 do
  //  TObject(EnstantiationList[I]).Free;
  EnstantiationList.Free;
end.
