unit Resource;

{
  Purpose: This library houses support types for reading complex resource streams.
    Its primary use is in Version.pas.
}

interface

uses
  SysUtils, Classes, Windows;

const
  soFromMarker = 3;

type
  EInvalidResource = class (Exception);

  TExtResStream = class (TCustomMemoryStream)
  private
    HResInfo: HRSRC;
    HGlobal: THandle;
    procedure Initialize(Instance: THandle; Name, ResType: PChar);
  public
    constructor Create(Instance: THandle; const ResName: string; ResType: PChar);
    constructor CreateFromID(Instance: THandle; ResID: Integer; ResType: PChar);
    destructor Destroy; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    procedure SetPointer(Ptr: Pointer; Size: LongInt);
  end;

  TExtResSubStream = class (TCustomMemoryStream)
  public
    function Write(const Buffer; Count: Longint): Longint; override;
    function ReadKey : String;
    procedure ReadPadding;
    procedure SetPointer(Ptr: Pointer; Size: LongInt);
  end;

implementation

 //-- TExtResStream ------------------------------------------------------------

constructor TExtResStream.Create(Instance: THandle; const ResName: string;
  ResType: PChar);
begin
  inherited Create;
  Initialize(Instance, PChar(ResName), ResType);
end;

constructor TExtResStream.CreateFromID(Instance: THandle; ResID: Integer;
  ResType: PChar);
begin
  inherited Create;
  Initialize(Instance, PChar(ResID), ResType);
end;

procedure TExtResStream.Initialize(Instance: THandle; Name, ResType: PChar);

  procedure Error;
  var
    OutName, OutType : String;
  begin
    If Integer(Name) shr 16 = 0 then
      OutName := '#' + IntToStr(Integer(Name))
    else
      OutName := Name;

    If Integer(ResType) shr 16 = 0 then
      OutType := '#' + IntToStr(Integer(ResType))
    else
      OutType := ResType;

    raise EResNotFound.CreateFmt('Resource %s of type %s not found',[OutName,OutType]);
  end;

begin
  HResInfo := FindResource(Instance, Name, ResType);
  if HResInfo = 0 then Error;
  HGlobal := LoadResource(Instance, HResInfo);
  if HGlobal = 0 then Error;
  SetPointer(LockResource(HGlobal), SizeOfResource(Instance, HResInfo));
end;

destructor TExtResStream.Destroy;
begin
  UnlockResource(HGlobal);
  FreeResource(HGlobal);
  FreeResource(HResInfo);
  inherited Destroy;
end;

function TExtResStream.Write(const Buffer; Count: Longint): Longint;
begin
  raise EStreamError.Create('Can''t write resource streams');
end;

procedure TExtResStream.SetPointer(Ptr: Pointer; Size: LongInt);
begin
  inherited SetPointer(Ptr, Size);
end;

 //-- TExtResSubStream ---------------------------------------------------------

function TExtResSubStream.Write(const Buffer; Count: Longint): Longint;
begin
  Result := 0;
end;

function TExtResSubStream.ReadKey : String;
var
  Buffer : PWideChar;
  PBuffer : PChar;
  T : Integer;
  C : WChar;
  OOdd : Boolean;
begin
  T := 8-(Position and $07);
  GetMem(Buffer,1024);
  Read(Buffer^,T);

  OOdd := (Position and $1) = 1;
  If not OOdd and (T >= 2) then
    C := PWideChar(PChar(Buffer) + (T - 2))[0]
  else if OOdd and (T >= 3) then
    C := PWideChar(PChar(Buffer) + (T - 3))[0]
  else
    C := #00;
  PBuffer := PChar(Buffer) + T;

  While (C <> #00) and (Position < Size) do
    begin
      Read(PBuffer^,4);
      If OOdd then C := PWideChar(PBuffer + 1)[0] else C := PWideChar(PBuffer + 2)[0];
      Inc(PBuffer,4);
    end;

  PWideChar(PBuffer)[0] := #00;
  Result := WideCharToString(Buffer);
  FreeMem(Buffer,1024);
end;

procedure TExtResSubStream.ReadPadding;
begin
  Seek(soFromCurrent,8-(Position and $07));
end;

procedure TExtResSubStream.SetPointer(Ptr: Pointer; Size: LongInt);
begin
  inherited SetPointer(Ptr, Size);
end;

end.
