unit Util;

interface

uses
  Classes, Windows, ActiveX;

type
  TMemoryBlock = class (TObject)
    Size : Integer;
    Data : Pointer;
    constructor Create(ASize : Integer);
    constructor CreateIndirect(ASize : Integer; AData : Pointer);
    destructor Destroy; override;
  end;

  TMemoryBlockArray = Array[0..0] of TMemoryBlock;
  PMemoryBlockArray = ^TMemoryBlockArray;

  TFileQueueThread = class;
  TThreadedFileQueue = class (TStream)
  private
    FBlockCount : Integer;         // Total number of blocks (loaded or not)
    FBlocks : PMemoryBlockArray;   // The array of blocks
    FBlockInPtr : Integer;         // The current incoming block
    FBlockOutPtr : Integer;        // The current outgoing block
    FBlockDataPtr : Integer;       // Pointer inside FBlockOutPtr
    FBlockSize : Integer;          // User defined size of a block
    FPosition : Integer;
    FFileName : String;
    ReachedEOF : Boolean;
    Thread : TFileQueueThread;
    procedure DoThreadTerminate(Sender : TObject);
  public
    constructor Create(FileName : String);
    destructor Destroy; override;
     // TStream
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
     //
    procedure LoadFromStream(Stream : TStream);
    property BlockSize : Integer read FBlockSize write FBlockSize;
  end;

  TFileQueueThread = class (TThread)
  public
    FileName : String;
    Queue : TThreadedFileQueue;
    procedure Execute; override;
  end;

procedure RaiseKnownWin32Error(RetVal : Integer);
function Win32CheckKnown(Error : Integer) : Integer;
procedure PopulateStringList(Source : String; Dest : TStringList);
procedure FlushObjectList(List : TList);
function CreateSafeArrayFromStream(S : TStream) : PSafeArray;
procedure DoSound(Name : String);
function StrToIP(S : String) : Cardinal;
function IPToStr(I : Cardinal) : String;
function LoadUserName : String;
function ExtractString(FuncName, StrName : String; const Data : Array of Byte; var I : Integer) : String;
function HexStrToInt(S : String) : Integer;

implementation

uses
  SysUtils, MMSystem;

 //-- TMemoryBlock -------------------------------------------------------------

constructor TMemoryBlock.Create(ASize : Integer);
begin
  inherited Create;
  Size := ASize;
  GetMem(Data, ASize);
end;

constructor TMemoryBlock.CreateIndirect(ASize : Integer; AData : Pointer);
begin
  inherited Create;
  Size := ASize;
  Data := AData;
end;

destructor TMemoryBlock.Destroy;
begin
  FreeMem(Data, Size);
  inherited Destroy;
end;

 //-- TThreadedFileQueue -------------------------------------------------------

constructor TThreadedFileQueue.Create(FileName : String);
begin
  if not FileExists(FileName) then
    raise Exception.CreateFmt('File %s does not exist', [FileName]);

  inherited Create;
  FFileName := FileName;
  FBlockSize := 1024;

  Thread := TFileQueueThread.Create(True);
  Thread.FreeOnTerminate := True;
  Thread.OnTerminate := DoThreadTerminate;
  Thread.FileName := FileName;
  Thread.Queue := Self;
  Thread.Resume;
end;

procedure TThreadedFileQueue.DoThreadTerminate(Sender : TObject);
begin
  Thread := Nil;
end;

destructor TThreadedFileQueue.Destroy;
var
  I : Integer;
begin
  if FBlockCount > 0 then
    begin
      For I := 0 to FBlockCount-1 do
        TMemoryBlock(FBlocks^[I]).Free;
      FreeMem(FBlocks, FBlockCount * SizeOf(TMemoryBlock));
    end;

  inherited Destroy;
end;

function TThreadedFileQueue.Read(var Buffer; Count: Longint): Longint;
var
  OutPtr, CopyCount : Integer;
begin
  OutPtr := 0;
  Result := 0;
  while Count > 0 do
    begin
      if FBlockOutPtr < FBlockInPtr then
        begin
           // If we're copying a block of data smaller than the current block...
          if Count < FBlocks^[FBlockOutPtr].Size - FBlockDataPtr then
            CopyCount := Count
           // ...copy as much data as we can from the current block.
          else
            CopyCount := FBlocks^[FBlockOutPtr].Size - FBlockDataPtr;

           // Move the data.
          if @Buffer <> Nil then
            Move((PChar(FBlocks^[FBlockOutPtr].Data)+FBlockDataPtr)^,
                 (PChar(@Buffer)+OutPtr)^,
                 CopyCount);
          Inc(FPosition, CopyCount);
          Dec(Count, CopyCount);
          Inc(OutPtr, CopyCount);
          Inc(FBlockDataPtr, CopyCount);
          Inc(Result, CopyCount);
           // Should never be ">"
          Assert(FBlockDataPtr <= FBlocks^[FBlockOutPtr].Size);
          if FBlockDataPtr = FBlocks^[FBlockOutPtr].Size then
            begin
              FBlockDataPtr := 0;
               // Toss blocks as we're done with them.
              with FBlocks^[FBlockOutPtr] do
                begin
                  FreeMem(Data, Size);
                  Data := Nil;
                  Size := 0;
                end;
              Inc(FBlockOutPtr);
            end;
        end
      else if ReachedEOF and (FBlockOutPtr >= FBlockInPtr) then
        Count := 0;
    end;
end;

function TThreadedFileQueue.Write(const Buffer; Count: Longint): Longint;
begin
  raise Exception.Create('TThreadedFileQueue.Write is not a valid call');
end;

function TThreadedFileQueue.Seek(Offset: Longint; Origin: Word): Longint;
var
  N : Pointer;
begin
   // Warning!  The use of N is wierd.  The problem is that I need to get a
   // nil into Buffer (not to what buffer points to).  This should do it, but
   // it's kind of unorthodox.
  N := Nil;
  case Origin of
    soFromCurrent :
      if Offset > 0 then
        Read(N^, Offset);
    soFromBeginning :
      if Offset - FPosition > 0 then
        Read(N^, Offset - FPosition)
      else if Offset - FPosition < 0 then
        raise Exception.Create('Cannot seek backwards in TThreadedFileQueue');
  else
    raise Exception.Create('Cannot seek backwards in TThreadedFileQueue');
  end;
  Result := FPosition;
end;

procedure Delay;
var
  I : Integer;
begin
  For I := 0 to 100000 do;
end;

procedure TThreadedFileQueue.LoadFromStream(Stream : TStream);
var
  Buffer : Pointer;
  Done : Boolean;
  ReadSize : Integer;
  S,  S2 : Integer;
begin
  ReachedEOF := False;

   // Load the data in from the stream one block at a time.
  FBlockCount := (Stream.Size + FBlockSize - 1) div FBlockSize;
  GetMem(FBlocks, FBlockCount * SizeOf(TMemoryBlock));
  FillChar(FBlocks^, FBlockCount * SizeOf(TMemoryBlock), 0);

  S := Stream.Size;
  S2 := 0;

  FBlockInPtr := 0;
  Done := False;
  GetMem(Buffer, FBlockSize);
  try
    while not Done do
      begin
//        Delay;
        ReadSize := Stream.Read(Buffer^, FBlockSize);
        Inc(S2, ReadSize);
         // No data left--just quit
        if ReadSize = 0 then
          Done := True
         // Full block read.  Just copy it over.
        else if ReadSize = FBlockSize then
          begin
            FBlocks^[FBlockInPtr] :=
              TMemoryBlock.CreateIndirect(ReadSize, Buffer);
            GetMem(Buffer, FBlockSize);
            Inc(FBlockInPtr);
          end
         // Partial block copied.  Create scaled down block and copy.
        else
          begin
            FBlocks^[FBlockInPtr] := TMemoryBlock.Create(ReadSize);
            Move(Buffer^, FBlocks^[FBlockInPtr].Data^, ReadSize);
            Inc(FBlockInPtr);
            Done := True;
          end;
      end;
    if S <> S2 then
      MessageBox(0, PChar(Format('Read descr %d, %d', [S, S2])), 'Error', MB_OK or MB_ICONERROR);
  finally
    ReachedEOF := True;  
    if Buffer <> Nil then FreeMem(Buffer, FBlockSize);
  end;
end;

 //-- TFileQueueThread ---------------------------------------------------------

procedure TFileQueueThread.Execute;
var
  F : TFileStream;
begin
  try
    F := TFileStream.Create(FileName, fmOpenRead);
    try
      Queue.LoadFromStream(F);
    finally
      F.Free;
    end;
  except
    on E: Exception do
      MessageBox(0, PChar(E.Message), 'Exception', MB_OK or MB_ICONERROR);
  end;
end;

 //-- Functions and Procedures -------------------------------------------------

procedure RaiseKnownWin32Error(RetVal : Integer);
var
  Error: EWin32Error;
begin
  if RetVal <> ERROR_SUCCESS then
    Error := EWin32Error.CreateFmt('Win 32 Error %d:%s', [RetVal,
      SysErrorMessage(RetVal)])
  else
    Error := EWin32Error.Create('Unknown Win 32 error');
  Error.ErrorCode := RetVal;
  raise Error;
end;

function Win32CheckKnown(Error : Integer) : Integer;
begin
  if Error <> ERROR_SUCCESS then RaiseKnownWin32Error(Error);
  Result := Error;
end;

procedure PopulateStringList(Source : String; Dest : TStringList);
var
  ParseState : (psElement, psBreak, psDone);
  C : Char;
  I : Integer;
  Start : Integer;
begin
  I := 1;
  Start := 1;
  ParseState := psElement;
  While ParseState <> psDone do
    begin
      C := Source[I];
      Case ParseState of
        psElement:
          Case C of
            #00 :
              begin
                Dest.Add(Copy(Source,Start,I-Start));
                ParseState := psDone;
              end;
            ',' :
              begin
                Dest.Add(Trim(Copy(Source,Start,I-Start)));
                ParseState := psBreak;
                Inc(I);
              end;
            else
              Inc(I);
          end;
        psBreak:
          Case C of
            #00 :
              ParseState := psDone;
            ' ' :
              Inc(I);
            else
              begin
                Start := I;
                ParseState := psElement;
              end;
          end;
      end;
    end;
end;

{
function Min(A, B : Integer) : Integer; Assembler
asm
  mov eax, A
  cmp eax, B
  jl @A_IS_LESS
  mov eax, B
@A_IS_LESS:
end;

function Max(A, B : Integer) : Integer; Assembler
asm
  mov eax, A
  cmp eax, B
  jg @A_IS_MORE
  mov eax, B
@A_IS_MORE:
end;
}

procedure FlushObjectList(List : TList);
var
  I : Integer;
begin
   // Frees all of the items in a TList, expecting them
   // to be TObject based.
  if List = Nil then
    Exit;
  For I := 0 to List.Count-1 do
    TObject(List[I]).Free;
  List.Clear;
end;

function CreateSafeArrayFromStream(S : TStream) : PSafeArray;
var
  Bounds : Array[0..0] of TSafeArrayBound;
begin
  Bounds[0].cElements := S.Size;
  Bounds[0].lLBound := 0;
  Result := SafeArrayCreate(varByte, 1, Bounds);
  try
    S.Seek(0, soFromBeginning);
    S.Read(Result^.pvData^, S.Size);
  except
    SafeArrayDestroy(Result);
    raise;
  end;
end;

procedure DoSound(Name : String);
begin
  PlaySound(PChar(Name),0,SND_APPLICATION or SND_ALIAS or
                          SND_ASYNC or SND_NODEFAULT);
end;

function StrToIP(S : String) : Cardinal;
var
  C : Char;
  I, Start, ByteCounter : Integer;
begin
  I := 1;
  Start := 1;
  ByteCounter := 0;
  Result := 0;
  repeat
    C := S[I];
    if C in ['.', #00] then
      begin
        Result := Result or (Cardinal(StrToInt(Copy(S, Start, I-Start))) shl (8*ByteCounter));
        Start := I+1;
        Inc(ByteCounter);
      end;
    Inc(I);
  until (C = #00) or (ByteCounter >= 4);

  if ByteCounter < 4 then
    raise Exception.Create('Invalid IP address ' + S);
end;

function IPToStr(I : Cardinal) : String;
begin
  Result := Format('%d.%d.%d.%d', [I and $FF, (I shr 8) and $FF,
                                   (I shr 16) and $FF, (I shr 24) and $FF]);
end;

function LoadUserName : String;
var
  Size : DWORD;
  Res : BOOL;
begin
  Size := 1024;
  SetLength(Result, Size);
  Res := GetUserName(PChar(Result), Size);
  SetLength(Result, Size); // Either hone down string or expand it to fit
  if not Res then
    begin
      Res := GetUserName(PChar(Result), Size);
      if not Res then
        RaiseKnownWin32Error(GetLastError);
    end;
end;

const
  SInvalidPacket = 'Invalid %s packet.  Expecting %s';

function ExtractString(FuncName, StrName : String; const Data : Array of Byte; var I : Integer) : String;
var
  S : Integer;
begin
  S := I;
  while (I < Length(Data)) and (Data[I] <> 0) do Inc(I);
  if I >= Length(Data) then raise Exception.CreateFmt(SInvalidPacket, [FuncName, StrName]);
  SetLength(Result, I-S);
  Move(Data[S], Result[1], I-S);
  Inc(I);
end;

function HexStrToInt(S : String) : Integer;
var
  I : Integer;
begin
  Result := 0;
  for I := 1 to Length(S) do
    begin
      Result := Result shl 4;
      if S[I] in ['0'..'9'] then
        Inc(Result, Ord(S[I]) - Ord('0'))
      else if S[I] in ['A'..'F'] then
        Inc(Result, 10 + Ord(S[I]) - Ord('A'))
      else
        raise Exception.Create('Invalid Hex number');
    end;
end;

end.
