unit Options;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, XMLParser;

type
  TfrmOptions = class(TForm)
    btnOK: TButton;
    btnCancel: TButton;
    btnReload: TButton;
    ScrollBox1: TScrollBox;
    procedure btnReloadClick(Sender: TObject);
    procedure btnOKClick(Sender: TObject);
  private
    { Private declarations }
    RequiredFiles: TStringList;
  public
    { Public declarations }
    AnimBkColor : TColor;
    InsideUOData : TXMLParser;
    destructor Destroy; override;
    procedure SetPath(ID : Integer; const Path : String);
    function GetPath(const Name : String) : String;
    function VerifyFile(const FileName : String) : Boolean;
    procedure ReloadPaths(Overwrite : Boolean);
    procedure Startup;
    procedure Save;
  end;

var
  frmOptions: TfrmOptions;

function StrLeft(const S: String; const Len: Integer): String;

implementation

{$R *.DFM}

uses
  Registry;

function StrLeft(const S: String; const Len: Integer): String;
var
  i : Integer;
  SLen : Integer;
begin
  SLen := Length(S);
  SetLength(Result, Len);
  for i := 1 to Len do
    begin
      if(i <= SLen) then
        Result[i] := S[i]
      else
        Result[i] := ' ';
    end;
end;

function GetMangledName(const Prefix : String; const Name : String) : String;
var
  I : Integer;
begin
  Result := Name;
  For I := 1 to Length(Name) do
    if Result[I] = '.' then Result[I] := '_';
  Result := Prefix + Result;
end;

destructor TfrmOptions.Destroy;
begin
  RequiredFiles.Free;
  InsideUOData.Free;
  inherited Destroy;
end;

procedure TfrmOptions.SetPath(ID : Integer; const Path : String);
var
  MName : String;
  Comp : TComponent;
begin
  MName := GetMangledName('edt', RequiredFiles[ID]);
  Comp := FindComponent(MName);
  if Comp = Nil then
    raise Exception.CreateFmt('Internal error: Could not find path %s', [MName]);
  with Comp as TEdit do
    Text := Path;
end;

function TfrmOptions.GetPath(const Name : String) : String;
var
  Comp : TComponent;
  MName : String;
begin
  MName := GetMangledName('edt', Name);
  Comp := FindComponent(MName);
  if Comp = Nil then
    raise Exception.CreateFmt('Internal error: Could not find component %s.  Does the <Files> tag in InsideUO.xml include %s?', [MName, Name]);
  with Comp as TEdit do
    Result := Text;
end;

const
  UORegKey = '\SOFTWARE\Origin Worlds Online\Ultima Online\1.0';
  UOExePathValue = 'ExePath';
  UOInstCDPathValue = 'InstCDPath';
  ErrorLoadingRegPaths =
        'There was a problem loading the paths.  This may be caused by ' +
        'an improper installation of UO or it may be an NT registry ' +
        'issue.  Either correct the problem, or enter the paths manually ' +
        'in View:Options.';
  WarningCaption = 'Warning';
  WarningVitalFilesMissing =
        'Some or all UO files could not be found.  Please check View:Options ' +
        'for misplaced files.';

procedure TfrmOptions.ReloadPaths(Overwrite : Boolean);
var
  Reg : TRegistry;
  MachinePath,
  CDPath : String;
  FileFindError : Boolean;
  I : Integer;
begin
  Reg := TRegistry.Create;
  try
    Reg.RootKey := HKEY_LOCAL_MACHINE;
    if not Reg.OpenKey(UORegKey, False) then
      raise Exception.Create(ErrorLoadingRegPaths);

    MachinePath := ExtractFilePath(Reg.ReadString(UOExePathValue));
    CDPath := Reg.ReadString(UOInstCDPathValue) + 'Client\';
  finally
    Reg.Free;
  end;

  FileFindError := False;
  For I := 0 to RequiredFiles.Count-1 do
    begin
      if Overwrite or (GetPath(RequiredFiles[I]) = '') then
        begin
          if FileExists(MachinePath + RequiredFiles[I]) then
            SetPath(I, MachinePath + RequiredFiles[I])
          else if FileExists(CDPath + RequiredFiles[I]) then
            SetPath(I, CDPath + RequiredFiles[I])
          else
            begin
              if(Integer(RequiredFiles.Objects[I]) = 1) then
                FileFindError := True;
              SetPath(I, '');
            end;
        end;
    end;

  if FileFindError then
    Application.MessageBox(WarningVitalFilesMissing, WarningCaption, MB_OK);
end;

procedure TfrmOptions.btnReloadClick(Sender: TObject);
begin
  ReloadPaths(True);
end;

procedure TfrmOptions.btnOKClick(Sender: TObject);
begin
  Save;
end;

procedure TfrmOptions.Startup;
var
  Reg : TRegistry;
  I : Integer;
  Str : String;
  CompLabel : TLabel;
  CompEdit : TEdit;
  Node : TXMLNode;
begin
  RequiredFiles := TStringList.Create;
//  DatFileName := ExtractFilePath(ParamStr(0)) + DatFileName;
  InsideUOData := TXMLParser.Create('InsideUO.xml');

  Node := InsideUOData.GetNode('InsideUO/Files');
  for I := 0 to Node.Count-1 do
    begin
      if Node.Items[I].TagName = 'FILE' then
        begin
          if Node.Items[I].GetAttributeValue('VITAL') = 'FALSE' then
            RequiredFiles.AddObject(Node.Items[I].Data, TObject(0))
          else
            RequiredFiles.AddObject(Node.Items[I].Data, TObject(1));
        end;
    end;

   // Defaults
  AnimBkColor := clBtnFace;

  for I := 0 to RequiredFiles.Count-1 do
    begin
      CompLabel := TLabel.Create( Self );
      CompLabel.Parent := ScrollBox1;
      CompLabel.Name := GetMangledName('lbl', RequiredFiles[I]);
      CompLabel.Top := 8 + 24 * I;
      CompLabel.Left := 8;
      CompLabel.Caption := RequiredFiles[I];

      CompEdit := TEdit.Create( Self );
      CompEdit.Parent := ScrollBox1;
      CompEdit.Name := GetMangledName('edt', RequiredFiles[I]);
      CompEdit.Top := 4 + 24 * I;
      CompEdit.Left := 100;
      CompEdit.Width := 280;
      CompEdit.Text := '';
    end;

  Reg := TRegistry.Create;
  try
    if not Reg.OpenKey('\Software\BlackSquad\InsideUO', False) then
      ReloadPaths(True)
    else
      begin
        For I := 0 to RequiredFiles.Count-1 do
          try
            Str := Reg.ReadString(RequiredFiles[I]);
            SetPath(I, Str);
          except
          end;

        ReloadPaths(False);

        if Reg.ValueExists('AnimBkColor') then
          AnimBkColor := Reg.ReadInteger('AnimBkColor');
      end;
  finally
    Reg.Free;
  end;
end;

procedure TfrmOptions.Save;
var
  Reg : TRegistry;
  I : Integer;
  Str : String;
begin
  Reg := TRegistry.Create;
  try
    Reg.OpenKey('\Software\BlackSquad\InsideUO', True);
    try
      For I := 0 to RequiredFiles.Count-1 do
        begin
          Str := GetPath(RequiredFiles[I]);
          Reg.WriteString(RequiredFiles[I], Str);
          Reg.WriteInteger('AnimBkColor',AnimBkColor);
        end;
    finally
      Reg.CloseKey;
    end;
  finally
    Reg.Free;
  end;
end;

function TfrmOptions.VerifyFile(const FileName : String) : Boolean;
var
  Text : String;
begin
  if not FileExists(GetPath(FileName)) then
    begin
      Text := 'Configure InsideUO to point to an existing copy of ' + FileName + ' before accessing this option.';
      Application.MessageBox(PChar(Text), 'Error', MB_OK );
      result := false;
    end
  else
    begin
      result := true;
    end;
end;

end.
