unit IndexedFile;

{
Most of the files in Ultima Online have an Index File which is made up of a
fairly consistant structure and which points to locations in a Data File.
The TIndexedFile handles these while at the same time overlaying them with
a patch file.
}

interface

uses
  Classes, SysUtils;

const
  CACHE_SIZE = 50;

type
  TIndexedData = class (TObject)
  private
     // Only used by the caching
    CacheIdx,
    OrderedIdx : Integer;
  public
    Data : TMemoryStream;
    Handle : Integer;
    Tag : Integer;
    Source : TStream;
    destructor Destroy; override;
  end;

  TIndexedFile = class (TObject)
  private
    FData,
    FPatch : TStream;
    FIndexList : TList;

    Cache : Array[0..CACHE_SIZE-1] of TIndexedData;
    CachePtr : Integer;
    OrderedCache : TList;

    procedure NeedCacheIndex(I : Integer);
  protected
    function GetCount : Integer;
    function GetData(AIndex : Integer) : TIndexedData;
  public
    constructor Create(AIndexFileName, ADataFileName : String;
                       APatchFile : TStream;
                       ADataPatchID : Integer;
                       RemoveGaps : Boolean);
    destructor Destroy; override;
    function FindHandle(AHandle : Integer; var Location : Integer) : Boolean; overload;
    function FindHandle(AHandle : Integer; Start : Integer; var Location : Integer) : Boolean; overload;
    procedure GetRangeInfo(St, Ed : Integer; var First : Integer; var Count : Integer);
    procedure GetEntryInfo(Index : Integer; var Handle : Integer; var Tag : Integer; var Source : TStream);
    property Count : Integer read GetCount;
    property Items[Index : Integer] : TIndexedData read GetData; default;
  end;

implementation

uses
  Util, Dialogs;

type
 {$A-}
  TIndexRec = record
    Lookup,
    Length,
    Tag : Integer;
  end;

  TPatchIndex = record
    FileID : Integer;
    Handle : Integer;
    Lookup : Integer;
    Length : Integer;
    Tag : Integer;
  end;
 {$A+}

  TIndex = class (TObject)
    Handle,
    Lookup,
    Length,
    Tag : Integer;
    Stream : TStream;
  end;

 //-- TIndexedData -------------------------------------------------------------

destructor TIndexedData.Destroy;
begin
  Data.Free;
  inherited Destroy;
end;

 //-- TIndexedFile -------------------------------------------------------------

constructor TIndexedFile.Create(AIndexFileName, ADataFileName : String;
                                APatchFile : TStream;
                                ADataPatchID : Integer;
                                RemoveGaps : Boolean);
var
  IndexRec : TIndexRec;
  IndexFile : TFileStream;
  IndexStream : TMemoryStream;
  Index : TIndex;
  I, L : Integer;
  PatchCount : Integer;
  PatchIndex : TPatchIndex;
begin
  inherited Create;

  FData := TFileStream.Create(ADataFileName, fmOpenRead or fmShareDenyNone);
  FPatch := APatchFile;
  FIndexList := TList.Create;
  OrderedCache := TList.Create;

   // Load the contents of the index file into the index list
  I := 0;
  IndexStream := TMemoryStream.Create;
  try
     // Load the index file into a memory stream
    IndexFile := TFileStream.Create(AIndexFileName, fmOpenRead or fmShareDenyNone);
    try
      IndexStream.LoadFromStream(IndexFile);
    finally
      IndexFile.Free;
    end;

    while IndexStream.Position < IndexStream.Size do
      begin
        IndexStream.Read(IndexRec, SizeOf(IndexRec));
        // TODO: There may be issues in LBR with the fact that lookup 0 is not
        //       valid.
        if not RemoveGaps or ((IndexRec.Lookup >= 0) and (IndexRec.Length > 0)) then
          begin
            Index := TIndex.Create;
            Index.Handle := I;
            Index.Lookup := IndexRec.Lookup;
            Index.Length := IndexRec.Length;
            Index.Tag := IndexRec.Tag;
            Index.Stream := FData;
            FIndexList.Add(Index);
          end;
        Inc(I);
      end;
    FIndexList.Pack;
  finally
    IndexStream.Free;
  end;

   // Load the contents of the patch file into the index list
  if FPatch <> Nil then
    begin
      FPatch.Seek(0, soFromBeginning);
      FPatch.Read(PatchCount, SizeOf(PatchCount));
      for I := 0 to PatchCount-1 do
        begin
          FPatch.Read(PatchIndex, SizeOf(PatchIndex));
          if PatchIndex.FileID = ADataPatchID then
            begin
              if FindHandle(PatchIndex.Handle, L) then
                Index := TIndex(FIndexList[L])
              else
                begin
                  Index := TIndex.Create;
                  Index.Handle := PatchIndex.Handle;
                  FIndexList.Insert(L, Index);
                end;
              Index.Lookup := PatchIndex.Lookup;
              Index.Length := PatchIndex.Length;
              Index.Tag := PatchIndex.Tag;
              Index.Stream := FPatch;
            end;
        end;
    end;
end;

destructor TIndexedFile.Destroy;
begin
  FlushObjectList(OrderedCache);
  OrderedCache.Free;
  OrderedCache := Nil;
  FlushObjectList(FIndexList);
  FIndexList.Free;
  FIndexList := Nil;
  FData.Free;
  inherited Destroy;
end;

function TIndexedFile.GetCount : Integer;
begin
  if FIndexList = Nil then
    Result := 0
  else
    Result := FIndexList.Count;
end;

function TIndexedFile.FindHandle(AHandle : Integer; var Location : Integer) : Boolean;
begin
  Location := 0;
  while (Location < FIndexList.Count) and
        (TIndex(FIndexList[Location]).Handle < AHandle) do
    Inc(Location);
  Result := (Location < FIndexList.Count) and
            (TIndex(FIndexList[Location]).Handle = AHandle);
end;

function TIndexedFile.FindHandle(AHandle : Integer; Start : Integer; var Location : Integer) : Boolean;
begin
  Location := Start;
  while (Location < FIndexList.Count) and
        (TIndex(FIndexList[Location]).Handle < AHandle) do
    Inc(Location);
  Result := (Location < FIndexList.Count) and
            (TIndex(FIndexList[Location]).Handle = AHandle);
end;

procedure TIndexedFile.NeedCacheIndex(I : Integer);
var
  CacheNode : TIndexedData;
begin
  if (I > 0) and (I < CACHE_SIZE) then
    begin
      CacheNode := Cache[I];
      Move(Cache[0], Cache[1], I * SizeOf(Pointer));
      Cache[0] := CacheNode;
    end;
end;

procedure TIndexedFile.GetRangeInfo(St, Ed : Integer; var First : Integer; var Count : Integer);
begin
  First := 0;
  while (First < FIndexList.Count) and (TIndex(FIndexList[First]).Handle < St) do
    Inc(First);
  Count := First;
  while (Count < FIndexList.Count) and (TIndex(FIndexList[Count]).Handle < Ed) do
    Inc(Count);
  Count := Count - First;
end;

procedure TIndexedFile.GetEntryInfo(Index : Integer; var Handle : Integer; var Tag : Integer; var Source : TStream);
begin
  if (Index < 0) or (Index >= FIndexList.Count) then
    raise Exception.CreateFmt('Invalid entry request %d for indexed file', [Index]);
  Handle := TIndex(FIndexList[Index]).Handle;
  Tag := TIndex(FIndexList[Index]).Tag;
  Source := TIndex(FIndexList[Index]).Stream;
end;

function TIndexedFile.GetData(AIndex : Integer) : TIndexedData;
var
  Index : TIndex;
  I : Integer;
  IndexedData : TIndexedData;
begin
  if (AIndex < 0) or (AIndex >= FIndexList.Count) then
    raise Exception.CreateFmt('Invalid data request %d for indexed file', [AIndex]);
  Index := TIndex(FIndexList[AIndex]);
  if Index.Lookup < 0 then
    raise Exception.Create('Trying to access empty entry');

   // Find entry in the ordered list
  I := 0;
  while (I < OrderedCache.Count) and (TIndexedData(OrderedCache[I]).OrderedIdx < AIndex) do
    Inc(I);

  if (I < OrderedCache.Count) and
     (TIndexedData(OrderedCache[I]).OrderedIdx = AIndex) then
    begin
       // Entry is in the cache
      Result := TIndexedData(OrderedCache[I]);
       // Move it to the front
      NeedCacheIndex(Result.CacheIdx);
      Result.CacheIdx := 0;
    end
  else
    begin
       // Entry was not found in the cache, so create it
      Result := TIndexedData.Create;
      Result.CacheIdx := CachePtr;
      Result.OrderedIdx := AIndex;
      Result.Data := TMemoryStream.Create;
      Result.Data.Size := Index.Length;
      Index.Stream.Seek(Index.Lookup, soFromBeginning);
      Result.Data.CopyFrom(Index.Stream, Index.Length);
      Result.Handle := Index.Handle;
      Result.Tag := Index.Tag;
      Result.Source := Index.Stream;

       // Load into the orderd cache
      OrderedCache.Insert(I, Result);

       // Load into the drop-off cache
      IndexedData := Cache[CACHE_SIZE-1];
       // If and object dropped off the end of the cache, then get rid of it
      if IndexedData <> NIL then
        begin
           // Find the entry in the ordered list...
          I := 0;
          while (I < OrderedCache.Count) and
                (TIndexedData(OrderedCache[I]).OrderedIdx < IndexedData.OrderedIdx) do
            Inc(I);
           // ...and delete it
          Assert((I < OrderedCache.Count) and
                 (TIndexedData(OrderedCache[I]).OrderedIdx = IndexedData.OrderedIdx));
          OrderedCache.Delete(I);
           // Free the object
          IndexedData.Free;
        end;
       // Shift the drop of cache and stick the new entry in the front
      Move(Cache[0], Cache[1], (CACHE_SIZE-1) * SizeOf(Pointer));
      Cache[0] := Result;
    end;
end;

end.
